def modify_quadratize():
    import numpy as np

    import serums.models as smodels
    import gncpy.control as gcontrol
    from gncpy.dynamics.basic import IRobotCreate
    from carbs.guidance import ELQR

    d2r = np.pi / 180

    tt = 0  # starting time when calculating control
    dt = 0.01
    time_horizon = 90 * dt  # run for 90 timesteps

    # define starting and ending state for control calculation
    end_states = [
        np.array([1.5, 2, 90 * d2r]).reshape((3, 1)),
        np.array([0, 2.5, 90 * d2r]).reshape((3, 1)),
        np.array([-1.5, 2, 90 * d2r]).reshape((3, 1)),
    ]
    start_dist = smodels.GaussianMixture()
    start_dist.add_components(
        np.array([-1, -0.25, 90 * d2r]).reshape((-1, 1)),
        np.diag([0.25, 0.25, 1 * d2r]),
        0.5,
    )
    start_dist.add_components(
        np.array([1, -0.25, 90 * d2r]).reshape((-1, 1)),
        np.diag([0.25, 0.25, 1 * d2r]),
        0.5,
    )

    # define nominal control input
    u_nom = np.zeros((2, 1))

    # define dynamics
    # IRobot Create has a dt that can be set here or it can be set by the control
    # algorithm
    dynObj = IRobotCreate(wheel_separation=0.258, radius=0.335 / 2.0)

    # define Q and R weights for using standard cost function
    Q = np.diag([1500, 1000, 1000])
    R = 0.5 * np.eye(u_nom.size)

    # define modifications for quadratizing the cost function
    def quad_modifier(itr, tt, P, Q, R, q, r):
        rot_cost = 0.4
        # only modify if within the first 4 iterations
        if itr < 4:
            Q[-1, -1] = rot_cost
            q[-1] = -rot_cost * np.pi / 2

        return P, Q, R, q, r

    # create control obect
    singleELQR = gcontrol.ELQR(time_horizon=time_horizon)
    singleELQR.set_state_model(u_nom, dynObj=dynObj)
    singleELQR.dt = dt  # set here or within the dynamic object
    singleELQR.set_cost_model(Q=Q, R=R)

    # create guidance object
    safety_factor = 4
    y_ref = 0.9

    elqr = ELQR()
    elqr.set_control_model(singleELQR)
    elqr.set_cost_model(
        quad_modifier=quad_modifier,
        safety_factor=safety_factor,
        y_ref=y_ref,
    )

    # convert end locations into a distribution or "wayareas"
    end_dist = elqr.targets_to_wayareas(end_states)

    # calculate guidance
    (
        state_trajectories,
        costs,
        control_signals,
        fig,
        frame_list,
    ) = elqr.plan(
        tt,
        start_dist,
        end_dist,
        provide_details=True,
        show_animation=True,
        save_animation=True,
        plt_inds=[0, 1],
        ttl="Multi-Agent GM Based ELQR",
    )

    return frame_list


def elqr_ospa():
    import numpy as np

    import serums.models as smodels
    import gncpy.control as gcontrol
    from gncpy.dynamics.basic import IRobotCreate
    from carbs.guidance import ELQROSPA
    from serums.distances import SingleObjectDistance

    d2r = np.pi / 180

    tt = 0  # starting time when calculating control
    dt = 0.01
    time_horizon = 90 * dt  # run for 90 timesteps

    # define starting and ending state for control calculation
    end_states = np.array([[1.5, 2, 90 * d2r], [0, 2.5, 90 * d2r], [-1.5, 2, 90 * d2r]])
    start_states = np.array([[-1, -0.25, 90 * d2r], [1, -0.25, 90 * d2r]])

    # define nominal control input
    u_nom = np.zeros((2, 1))

    # define dynamics
    # IRobot Create has a dt that can be set here or it can be set by the control
    # algorithm
    dynObj = IRobotCreate(wheel_separation=0.258, radius=0.335 / 2.0)

    # define Q and R weights for using standard cost function
    Q = np.diag([1500, 1000, 1000])
    R = 0.5 * np.eye(u_nom.size)

    # define modifications for quadratizing the cost function
    def quad_modifier(itr, tt, P, Q, R, q, r):
        rot_cost = 0.4
        # only modify if within the first 4 iterations
        if itr < 4:
            Q[-1, -1] = rot_cost
            q[-1] = -rot_cost * np.pi / 2

        return P, Q, R, q, r

    # create control obect
    singleELQR = gcontrol.ELQR(time_horizon=time_horizon)
    singleELQR.set_state_model(u_nom, dynObj=dynObj)
    singleELQR.dt = dt  # set here or within the dynamic object
    singleELQR.set_cost_model(Q=Q, R=R)

    # create guidance object
    pos_inds = [0, 1]
    cutoff = 10

    elqr = ELQROSPA()
    elqr.set_control_model(singleELQR)
    elqr.set_cost_model(
        quad_modifier=quad_modifier,
        ospa_inds=pos_inds,
        ospa_cutoff=cutoff,
        ospa_method=SingleObjectDistance.EUCLIDEAN,
    )

    # calculate guidance
    (
        state_trajectories,
        costs,
        control_signals,
        fig,
        frame_list,
    ) = elqr.plan(
        tt,
        start_states,
        end_states,
        provide_details=True,
        show_animation=True,
        save_animation=True,
        plt_inds=pos_inds,
        ttl="Multi-Agent OSPA Based ELQR",
    )

    return frame_list


def elqr_ospa_obstacles():
    import matplotlib.pyplot as plt
    import numpy as np
    from matplotlib.patches import Circle

    import gncpy.control as gcontrol
    import gncpy.plotting as gplot
    from gncpy.dynamics.basic import IRobotCreate
    from carbs.guidance import ELQROSPA
    from serums.distances import SingleObjectDistance

    d2r = np.pi / 180

    tt = 0  # starting time when calculating control
    dt = 0.01
    time_horizon = 150 * dt  # run for 120 timesteps

    # define starting and ending state for control calculation
    end_states = np.array(
        [[1.5, 2, 90 * d2r], [0, 0.75, 90 * d2r], [-1.5, 2, 90 * d2r]]
    )
    start_states = np.array([[-1, -0.15, 90 * d2r], [1, -0.15, 90 * d2r]])

    # define nominal control input
    u_nom = np.zeros((2, 1))

    # define dynamics
    # IRobot Create has a dt that can be set here or it can be set by the control
    # algorithm
    dynObj = IRobotCreate(wheel_separation=0.258, radius=0.335 / 2.0)

    # define Q and R weights for using standard cost function
    Q = np.diag([1500, 1000, 1000])
    R = 0.4 * np.eye(u_nom.size)

    # define modifications for quadratizing the cost function
    def quad_modifier(itr, tt, P, Q, R, q, r):
        rot_cost = 0.4
        # only modify if within the first 2 iterations
        if itr < 2:
            Q[-1, -1] = rot_cost
            q[-1] = -rot_cost * np.pi / 2

        return P, Q, R, q, r

    # create control obect
    singleELQR = gcontrol.ELQR(time_horizon=time_horizon)
    singleELQR.set_state_model(u_nom, dynObj=dynObj)
    singleELQR.dt = dt  # set here or within the dynamic object
    singleELQR.set_cost_model(Q=Q, R=R)  # can also set the quad modifier here

    # define some circular obstacles with center pos and radius (x, y, radius)
    obstacles = np.array(
        [
            [0, -1.35, 0.2],
            [1.0, -0.5, 0.2],
            [-0.95, -0.5, 0.2],
            [-0.2, 0.3, 0.2],
            [0.8, 0.7, 0.2],
            [1.1, 2.0, 0.2],
            [-1.2, 0.8, 0.2],
            [-1.1, 2.1, 0.2],
            [-0.1, 1.6, 0.2],
            [-1.1, -1.9, 0.2],
            [(10 + np.sqrt(2)) / 10, (-15 - np.sqrt(2)) / 10, 0.2],
        ]
    )

    # define enviornment bounds for the robot
    bottom_left = np.array([-2, -3])
    top_right = np.array([2, 3])

    # define additional non-quadratic cost
    obs_factor = 1
    scale_factor = 1
    cost_args = (obstacles, obs_factor, scale_factor, bottom_left, top_right)

    def non_quadratic_cost(
        tt,
        state,
        ctrl_input,
        end_state,
        is_initial,
        is_final,
        _obstacles,
        _obs_factor,
        _scale_factor,
        _bottom_left,
        _top_right,
    ):
        cost = 0
        # cost for obstacles
        for obs in _obstacles:
            diff = state.ravel()[0:2] - obs[0:2]
            dist = np.sqrt(np.sum(diff * diff))
            # signed distance is negative if the robot is within the obstacle
            signed_dist = (dist - dynObj.radius) - obs[2]
            if signed_dist > 0:
                continue
            cost += _obs_factor * np.exp(-_scale_factor * signed_dist).item()

        # add cost for going out of bounds
        for ii, b in enumerate(_bottom_left):
            dist = (state[ii] - b) - dynObj.radius
            cost += _obs_factor * np.exp(-_scale_factor * dist).item()

        for ii, b in enumerate(_top_right):
            dist = (b - state[ii]) - dynObj.radius
            cost += _obs_factor * np.exp(-_scale_factor * dist).item()

        return cost

    # create guidance object
    pos_inds = [0, 1]
    cutoff = 10

    elqr = ELQROSPA()
    elqr.set_control_model(singleELQR)
    elqr.set_cost_model(
        quad_modifier=quad_modifier,
        ospa_inds=pos_inds,
        ospa_cutoff=cutoff,
        ospa_method=SingleObjectDistance.EUCLIDEAN,
        non_quad_fun=non_quadratic_cost,
        non_quad_weight=10,
    )

    # create figure with obstacles to plot animation on
    cmap = gplot.get_cmap(start_states.shape[0])
    fig = plt.figure()
    fig.add_subplot(1, 1, 1)
    fig.axes[0].set_aspect("equal", adjustable="box")
    fig.axes[0].set_xlim((bottom_left[0], top_right[0]))
    fig.axes[0].set_ylim((bottom_left[1], top_right[1]))
    for c_ind, start_state in enumerate(start_states):
        fig.axes[0].scatter(
            start_state[pos_inds[0]],
            start_state[pos_inds[1]],
            marker="o",
            color=cmap(c_ind),
            zorder=1000,
        )
    for obs in obstacles:
        c = Circle(obs[:2], radius=obs[2], color="k", zorder=1000)
        fig.axes[0].add_patch(c)
    plt_opts = gplot.init_plotting_opts(f_hndl=fig)
    gplot.set_title_label(
        fig, 0, plt_opts, ttl="Multi-Agent OSPA Based ELQR w/ Obstacles"
    )

    # calculate guidance
    (
        state_trajectories,
        costs,
        control_signals,
        fig,
        frame_list,
    ) = elqr.plan(
        tt,
        start_states,
        end_states,
        cost_args=cost_args,
        provide_details=True,
        show_animation=True,
        save_animation=True,
        plt_inds=pos_inds,
        cmap=cmap,
        fig=fig,
    )

    return frame_list


def run():
    import os

    print("Generating ELQR examples")
    fps = 10
    duration = int(1 / fps * 1e3)

    fout = os.path.join(os.path.dirname(__file__), "elqr_modify_quadratize.gif")
    if not os.path.isfile(fout):
        frame_list = modify_quadratize()

        frame_list[0].save(
            fout,
            save_all=True,
            append_images=frame_list[1:],
            duration=duration,  # convert s to ms
            loop=0,
        )

    fout = os.path.join(os.path.dirname(__file__), "elqr_ospa_cost.gif")
    if not os.path.isfile(fout):
        frame_list = elqr_ospa()

        frame_list[0].save(
            fout,
            save_all=True,
            append_images=frame_list[1:],
            duration=duration,  # convert s to ms
            loop=0,
        )

    fout = os.path.join(os.path.dirname(__file__), "elqr_ospa_cost_obstacles.gif")
    if not os.path.isfile(fout):
        frame_list = elqr_ospa_obstacles()

        frame_list[0].save(
            fout,
            save_all=True,
            append_images=frame_list[1:],
            duration=duration,  # convert s to ms
            loop=0,
        )


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    plt.close("all")

    run()

    plt.show()



import sys
import os




def _run_guidance():
    sys.path.insert(0, os.path.abspath("./examples/guidance"))
    import elqr

    elqr.run()

    sys.path.pop(0)


def run_examples():
    _run_guidance()



"""Implements general sampling methods."""
import numpy as np


def gibbs(in_costs, num_iters, rng=None):
    """Implements a Gibbs sampler.

    Notes
    -----
    Computationally efficient form of the Metropolis Hastings Markov Chain
    Monte Carlo algorithm, useful when direct sampling of the target
    multivariate distribution is difficult. Generates samples by selecting
    from a histogram constructed from the distribution and randomly sampling
    from the most likely bins in the histogram. This is based on the sampler in
    :cite:`Vo2017_AnEfficientImplementationoftheGeneralizedLabeledMultiBernoulliFilter`.

    Parameters
    ----------
    in_costs : N x M numpy array
        Cost matrix.
    num_iters : int
        Number of iterations to run.
    rng : numpy random generator, optional
        Random number generator to be used when sampling. The default is None
        which implies :code:`default_rng()`

    Returns
    -------
    assignments : M (max size) x N numpy array
        The unique entries from the sampling.
    costs : M x 1 numpy array
        Cost of each assignment.
    """
    if rng is None:
        rng = np.random.default_rng()

    # Determine size of cost matrix
    cost_size = np.shape(in_costs)

    # Initialize assignment and cost matrices
    assignments = np.zeros((num_iters, cost_size[0]))
    costs = np.zeros((num_iters, 1))

    # Initialize current solution
    cur_soln = np.arange(cost_size[0], 2 * cost_size[0])

    assignments[0] = cur_soln
    rows = np.arange(0, cost_size[0]).astype(int)
    costs[0] = np.sum(in_costs[rows, cur_soln])

    # Loop over all possible assignments and determine costs
    mask = np.ones(len(cur_soln), dtype=bool)
    for sol in range(1, num_iters):
        for var in range(0, cost_size[0]):
            if var > 0:
                mask[var - 1] = True
            mask[var] = False
            temp_samp = np.exp(-in_costs[var])
            temp_samp[cur_soln[mask]] = 0

            hist_in_array = np.zeros(temp_samp[temp_samp > 0].size + 1)
            csum = np.cumsum(temp_samp[temp_samp > 0].ravel())
            hist_in_array[1:] = csum / csum[-1]

            cur_soln[var] = np.digitize(rng.uniform(size=(1,1)), hist_in_array) - 1
            if np.any(temp_samp > 0):
                cur_soln[var] = np.nonzero(temp_samp > 0)[0][cur_soln[var]]

        mask[-1] = True
        assignments[sol] = cur_soln
        costs[sol] = np.sum(in_costs[rows, cur_soln])

    [assignments, I] = np.unique(assignments, return_index=True, axis=0)

    return assignments, costs[I]

# TODO: Multisensor Gibbs potentially
def mm_gibbs(in_costs, num_iters, rng=None):
    """Implements a minimally-Markovian Gibbs sampler.

        Notes
        -----
        Computationally efficient form of the Metropolis Hastings Markov Chain
        Monte Carlo algorithm, useful when direct sampling of the target
        multivariate distribution is difficult. Generates samples by selecting
        from a histogram constructed from the distribution and randomly sampling
        from the most likely bins in the histogram. This is based on the sampler in
        :cite:`Vo2017_AnEfficientImplementationoftheGeneralizedLabeledMultiBernoulliFilter`.

        Parameters
        ----------
        in_costs : S x N x M numpy array
            Cost matrix.
        num_iters : int
            Number of iterations to run.
        rng : numpy random generator, optional
            Random number generator to be used when sampling. The default is None
            which implies :code:`default_rng()`

        Returns
        -------
        assignments : M (max size) x N x S numpy array
            The unique entries from the sampling.
        costs : M x S numpy array
            Cost of each assignment for each sensor.
        """

    if rng is None:
        rng = np.random.default_rng()

    # Determine size of cost matrix
    cost_size = np.shape(in_costs)

    # Initialize assignment and cost matrices
    #assignments is iterations x sensors x solution
    assignments = np.zeros((num_iters, cost_size[0], cost_size[1]))
    costs = np.zeros((num_iters, 1))
    P_n = []
    Q_n = []
    #TODO: calculation for P_n and Q_n
    for var in range(0, cost_size[1]):
        prod_gamma_n = 1
        prod_v_n = 1
        for s in range(0, cost_size[0]):
            gamma_n = np.sum(in_costs[s, var][in_costs[s, var] != -1]) #third index needs to represent -1 vs non neg
            prod_gamma_n = prod_gamma_n*gamma_n
            prod_v_n = prod_v_n * np.sum(in_costs[s, var][in_costs[s, var] == -1]) # index where in_costs are -1
        P_n.append(prod_gamma_n/(prod_v_n + prod_gamma_n))
        Q_n.append(1 - P_n[var])

    # Initialize current solution
    costsum = 0
    rows = np.arange(0, cost_size[1]).astype(int)
    cur_soln = np.zeros((cost_size[0], len(np.arange(cost_size[1], 2*cost_size[1]))),dtype=int)
    for s in range(0, cost_size[0]):
        cur_soln[s, :] = np.arange(cost_size[1], 2 * cost_size[1])
        costsum += np.sum(in_costs[s, rows, np.arange(cost_size[1], 2 * cost_size[1]).astype(int)])

    assignments[0, :, :] = cur_soln
    costs[0] = costsum

    # Loop over all possible assignments and determine costs
    mask = np.ones(np.shape(cur_soln), dtype=bool)
    for sol in range(1, num_iters):
        costsum = 0
        for var in range(0, cost_size[1]):
            #TODO: write first categorical
            # sample markovian stationary distribution,
            # if greater than some probability as calculated above by pn,
            # then perform mm-gibbs, otherwise all values are set to -1.
            #
            bins = np.array([0.0, Q_n[var], 1.0])
            i_n = np.digitize(rng.uniform(size=(1, 1)), bins) - 1
            "i_n = Categorical ('+', '-', ), [Pn(Lambda(1:num_sensors)), Qn(Lambda(1:num_sensors)]"
            if i_n > 0: # i_n = "+"

                for s in range(0, cost_size[0]):
                    if var > 0:
                        mask[s, var - 1] = True
                    mask[s, var] = False
                temp_samp = np.exp(-in_costs[:, var, :])
                temp_samp[cur_soln[mask]] = 0
                for s in range(0, cost_size[0]):
                    hist_in_array = np.zeros(temp_samp[s][temp_samp[s] > 0].size + 1)
                    csum = np.cumsum(temp_samp[s][temp_samp[s] > 0].ravel())
                    hist_in_array[1:] = csum / csum[-1]
                    cur_soln[s, var] = np.digitize(rng.uniform(size=(1, 1)), hist_in_array) - 1
                    if np.any(temp_samp[s] > 0):
                        cur_soln[s, var] = np.nonzero(temp_samp > 0)[0][cur_soln[s, var]]

            else: # i_n = "-"
                cur_soln[:, var] = -1 * np.ones(1, cost_size[0])

        assignments[sol, :, :] = cur_soln
        costs[sol] = np.sum(in_costs[:, rows, cur_soln]) #some variant of this


    [assignments, I] = np.unique(assignments, return_index=True, axis=0)

    return assignments, costs[I]




import argparse
import subprocess
import sys

from pathlib import Path
from setuptools_scm import get_version


def get_active_branch_name():
    head_dir = Path(".") / ".git" / "HEAD"
    with head_dir.open("r") as f:
        content = f.read().splitlines()

    for line in content:
        if line[0:4] == "ref:":
            return line.partition("refs/heads/")[2]


def define_parser() -> argparse.ArgumentParser:
    p = argparse.ArgumentParser(
        description="Automatically create and push tag so CI/CD pipeline publishes a release"
    )

    choices = ("major", "minor", "patch")
    p.add_argument(
        "type",
        type=str,
        help="Type of release to perform",
        choices=choices,
    )

    p.add_argument(
        "-s",
        "--skip-increment",
        action="store_true",
        help="Flag indicating if incrementing the version should be skipped. "
        + "If this is passed then the type is irrelevant.",
    )

    default = "Automatic release"
    p.add_argument(
        "-m",
        "--message",
        type=str,
        default=default,
        help="Message for the release (Doesn't seem to be used in CI/CD). The default is: {:s}".format(
            default
        ),
    )

    return p


if __name__ == "__main__":
    args = define_parser().parse_args()

    version_parts = get_version(root='.', relative_to=__file__, version_scheme="no-guess-dev").split('.')
    major = int(version_parts[0])
    minor = int(version_parts[1])
    patch = int(version_parts[2])

    print("Current version: {:d}.{:d}.{:d}".format(major, minor, patch))

    cur_branch = get_active_branch_name()
    if cur_branch.lower() != "master":
        print(
            "WARN: Not on master branch ({:s}), checkout to master branch for release".format(
                cur_branch
            )
        )
        sys.exit(-1)

    if not args.skip_increment:
        if args.type == "major":
            major += 1
            minor = 0
            patch = 0
        elif args.type == "minor":
            minor += 1
            patch = 0
        elif args.type == "patch":
            patch += 1
        else:
            raise RuntimeError("Invalid type: {} should not be here".format(args.type))

    version_str = "v{:d}.{:d}.{:d}".format(major, minor, patch)
    if args.skip_increment:
        print("Skipping incrementing of version number!")
        print("Removing old version tag")
        cmd_str = f"git tag -d {version_str}"
        print(cmd_str)
        subprocess.run(cmd_str, shell=True)

    print("Releasing version: {:s}".format(version_str[1:]))

    cmd_str = "git tag -a {:s} -m '{:s}'".format(version_str, args.message)
    print(cmd_str)
    subprocess.run(cmd_str, shell=True)

    cmd_str = "git push origin {:s}".format(version_str)
    subprocess.run(cmd_str, shell=True)



"""A python package for Command & control (C2) of Autonomous Random finite set- (RFS-) Based Swarms.

This code is developed by the Laboratory for Autonomy, Guidance, navigation,
control, and Estimation Research (LAGER) at the University of Alabama.
"""



""" Implements helper functions for graph search algorithms.

This module contains the longer helper functions for the graph search
algorithms.
"""
import numpy as np


def assign_opt(dist_mat_in):
    """ Optimal assignment used by Murty's m-best assignment algorithm.

    This ports the implementation from
    `here <https://ba-tuong.vo-au.com/codes.html>`_.
    to python.

    Args:
        dist_mat_in (numpy array): Distance matrix to calculate assignment from

    Returns:
        tuple containing

            - assign (numpy array): Optimal assignment
            - cost (float): Cost of the assignment
    """

    dist_mat = dist_mat_in.copy()
    if len(dist_mat.shape) == 1:
        dist_mat = dist_mat.reshape((1, dist_mat.shape[0]))
    elif len(dist_mat.shape) > 2:
        raise RuntimeError('Distance matrix must have at most 2 dimensions')
    dist_mat_orig = dist_mat.copy()

    (n_rows, n_cols) = dist_mat.shape
    assign = -np.ones(n_rows, dtype=int)
    if (np.isfinite(dist_mat) & (dist_mat < 0)).any():
        raise RuntimeError('All matrix values must be non-negative')

    if np.isinf(dist_mat).all():
        return (assign, 0)

    # replace any infite values with a large number
    if np.isinf(dist_mat).any():
        max_finite = dist_mat[np.nonzero(np.isfinite(dist_mat))]
        if len(max_finite) == 0:
            return (assign, 0)

        max_finite = max_finite.max()
        if max_finite > 0:
            inf_val = 10 * max_finite * n_rows * n_cols
        else:
            inf_val = 10
        dist_mat[np.nonzero(np.isinf(dist_mat))] = inf_val

    covered_cols = [False] * n_cols
    covered_rows = [False] * n_rows
    star_mat = np.zeros((n_rows, n_cols), dtype=bool)
    prime_mat = np.zeros((n_rows, n_cols), dtype=bool)

    if n_rows <= n_cols:
        min_dim = n_rows

        # find min of each row and subract from each element in the row
        dist_mat = dist_mat - dist_mat.min(axis=1).reshape((n_rows, 1))

        # steps 1 and 2a
        for row in range(0, n_rows):
            for col in range(0, n_cols):
                if dist_mat[row, col] == 0:
                    if not covered_cols[col]:
                        star_mat[row, col] = True
                        covered_cols[col] = True
                        break
    else:
        min_dim = n_cols

        # find min of each column and subract from each element in the column
        dist_mat = dist_mat - dist_mat.min(axis=0)

        # steps 1 and 2a
        for col in range(0, n_cols):
            for row in range(0, n_rows):
                if dist_mat[row, col].squeeze() == 0:
                    if not covered_rows[row]:
                        star_mat[row, col] = True
                        covered_cols[col] = True
                        covered_rows[row] = True
                        break

        for row in range(0, n_rows):
            covered_rows[row] = False

    __aop_step2b(assign, dist_mat, star_mat, prime_mat, covered_cols,
                 covered_rows, n_rows, n_cols, min_dim)
    cost = __aop_comp_assign(assign, dist_mat_orig, n_rows)

    return (assign, cost)


def __aop_step2a(assign, dist_mat, star_mat, prime_mat, covered_cols,
                 covered_rows, n_rows, n_cols, min_dim):
    for col in range(0, n_cols):
        for row in range(0, n_rows):
            if star_mat[row, col]:
                covered_cols[col] = True
                break

    __aop_step2b(assign, dist_mat, star_mat, prime_mat, covered_cols,
                 covered_rows, n_rows, n_cols, min_dim)


def __aop_step2b(assign, dist_mat, star_mat, prime_mat, covered_cols,
                 covered_rows, n_rows, n_cols, min_dim):
    n_covered_cols = 0
    for v in covered_cols:
        if v:
            n_covered_cols += 1

    if n_covered_cols == min_dim:
        __aop_build_assign(assign, star_mat, n_rows, n_cols)

    else:
        __aop_step3(assign, dist_mat, star_mat,
                    prime_mat, covered_cols, covered_rows, n_rows,
                    n_cols, min_dim)


def __aop_step3(assign, dist_mat, star_mat, prime_mat, covered_cols,
                covered_rows, n_rows, n_cols, min_dim):
    z_found = True
    while z_found:
        z_found = False
        for col in range(0, n_cols):
            if not covered_cols[col]:
                for row in range(0, n_rows):
                    cond1 = not covered_rows[row]
                    cond2 = dist_mat[row, col] == 0
                    if cond1 and cond2:
                        prime_mat[row, col] = True

                        star_col = 0
                        for ii in range(0, n_cols):
                            if star_mat[row, ii]:
                                star_col = ii
                                break
                            else:
                                star_col = ii + 1

                        if star_col == n_cols:
                            __aop_step4(assign, dist_mat, star_mat,
                                        prime_mat, covered_cols, covered_rows,
                                        n_rows, n_cols, min_dim, row, col)
                            return
                        else:
                            covered_rows[row] = True
                            covered_cols[star_col] = False
                            z_found = True
                            break

    __aop_step5(assign, dist_mat, star_mat, prime_mat, covered_cols,
                covered_rows, n_rows, n_cols, min_dim)


def __aop_step4(assign, dist_mat, star_mat, prime_mat, covered_cols,
                covered_rows, n_rows, n_cols, min_dim, row, col):
    new_star_mat = star_mat.copy()

    new_star_mat[row, col] = True
    star_col = col
    star_row = 0
    for ii in range(0, n_rows):
        star_row = ii
        if star_mat[ii, star_col]:
            break
        else:
            star_row = ii + 1


    while star_row < n_rows:
        new_star_mat[star_row, star_col] = False

        prime_row = star_row
        prime_col = 0
        for ii in range(0, n_cols):
            prime_col = ii
            if prime_mat[prime_row, ii]:
                break
            else:
                prime_col = ii + 1

        new_star_mat[prime_row, prime_col] = True

        star_col = prime_col
        for ii in range(0, n_rows):
            if star_mat[ii, star_col]:
                star_row = ii
                break
            else:
                star_row = ii + 1

    for row in range(0, n_rows):
        for col in range(0, n_cols):
            prime_mat[row, col] = False
            star_mat[row, col] = new_star_mat[row, col]

    for row in range(0, n_rows):
        covered_rows[row] = False

    __aop_step2a(assign, dist_mat, star_mat, prime_mat, covered_cols,
                 covered_rows, n_rows, n_cols, min_dim)


def __aop_step5(assign, dist_mat, star_mat, prime_mat, covered_cols,
                covered_rows, n_rows, n_cols, min_dim):
    h = np.inf
    for row in range(0, n_rows):
        if not covered_rows[row]:
            for col in range(0, n_cols):
                if not covered_cols[col]:
                    v = dist_mat[row, col]
                    if v < h:
                        h = v

    for row in range(0, n_rows):
        if covered_rows[row]:
            for col in range(0, n_cols):
                dist_mat[row, col] += h

    for col in range(0, n_cols):
        if not covered_cols[col]:
            for row in range(0, n_rows):
                dist_mat[row, col] -= h

    __aop_step3(assign, dist_mat, star_mat, prime_mat, covered_cols,
                covered_rows, n_rows, n_cols, min_dim)


def __aop_build_assign(assign, star_mat, n_rows, n_cols):
    for row in range(0, n_rows):
        for col in range(0, n_cols):
            if star_mat[row, col]:
                assign[row] = col
                break


def __aop_comp_assign(assign, dist_mat, n_rows):
    cost = 0
    for row in range(0, n_rows):
        col = assign[row]
        if col >= 0:
            v = dist_mat[row, col]
            if np.isfinite(v):
                cost += v
            else:
                assign[row] = -1
    return cost


class AStarNode:
    """
        A node class for A* Pathfinding
        parent is parent of the current Node
        position is current position of the Node in the maze
        g is cost from start to current Node
        h is heuristic based estimated cost for current Node to end Node
        f is total cost of present node i.e. :  f = g + h
        x and y are Cartesian coordinates of the node
    """

    def __init__(self, parent=None, position=None):
        self.parent = parent
        self.position = position

        self.g = 0
        self.h = 0
        self.f = 0
        self.x = 0
        self.y = 0

    def __eq__(self, other):
        return self.position == other.position


def astar_return_path(current_node, maze):
    path = []
    no_rows, no_columns = np.shape(maze)

    # here we create the initialized result maze with -1 in every position
    result = [[-1 for i in range(no_columns)] for j in range(no_rows)]
    current = current_node
    total_cost = []
    while current is not None:
        path.append(current.position)
        total_cost.append(current.h)
        current = current.parent

    # Return reversed path as we need to show from start to end path
    path = path[::-1]
    start_value = 0

    # we update the path of start to end found by A-star serch with every step
    # incremented by 1
    for i in range(len(path)):
        result[path[i][0]][path[i][1]] = start_value
        start_value += 1

    inds = np.argwhere(np.array(result) >= 0)
    vals = [result[r][c] for r, c in inds]
    s_inds = np.argsort(vals)
    vals = [vals[ii] for ii in s_inds]

    return (tuple([(inds[ii][0], inds[ii][1]) for ii in s_inds]), max(total_cost))



"""Implements RFS guidance algorithms.

This module contains the classes and data structures
for RFS guidance related algorithms.
"""
import io
import numpy as np
import scipy.linalg as la
import matplotlib.pyplot as plt
from PIL import Image
from copy import deepcopy
from warnings import warn
from scipy.optimize import linear_sum_assignment

import gncpy.control as gcontrol
import gncpy.plotting as gplot
import serums.models as smodels
from serums.distances import calculate_ospa
from serums.enums import SingleObjectDistance


def gaussian_density_cost(state_dist, goal_dist, safety_factor, y_ref):
    r"""Implements a GM density based cost function.

    Notes
    -----
    Implements the following cost function based on the difference between
    Gaussian mixtures, with additional terms to improve convergence when
    far from the targets.

    .. math::
        J &= \sum_{k=1}^{T} 10 N_g\sigma_{g,max}^2 \left( \sum_{j=1}^{N_g}
                \sum_{i=1}^{N_g} w_{g,k}^{(j)} w_{g,k}^{(i)}
                \mathcal{N}( \mathbf{m}^{(j)}_{g,k}; \mathbf{m}^{(i)}_{g,k},
                P^{(j)}_{g, k} + P^{(i)}_{g, k} ) \right. \\
            &- \left. 20 \sigma_{d, max}^2 N_d \sum_{j=1}^{N_d} \sum_{i=1}^{N_g}
                w_{d,k}^{(j)} w_{g,k}^{(i)} \mathcal{N}(
                \mathbf{m}^{(j)}_{d, k}; \mathbf{m}^{(i)}_{g, k},
                P^{(j)}_{d, k} + P^{(i)}_{g, k} ) \right) \\
            &+ \sum_{j=1}^{N_d} \sum_{i=1}^{N_d} w_{d,k}^{(j)}
                w_{d,k}^{(i)} \mathcal{N}( \mathbf{m}^{(j)}_{d,k};
                \mathbf{m}^{(i)}_{d,k}, P^{(j)}_{d, k} + P^{(i)}_{d, k} ) \\
            &+ \alpha \sum_{j=1}^{N_d} \sum_{i=1}^{N_g} w_{d,k}^{(j)}
                w_{g,k}^{(i)} \ln{\mathcal{N}( \mathbf{m}^{(j)}_{d,k};
                \mathbf{m}^{(i)}_{g,k}, P^{(j)}_{d, k} + P^{(i)}_{g, k} )}

    Parameters
    ----------
    state_dist : :class:`serums.models.GaussianMixture`
        Initial state distribution.
    goal_dist : :class:`serums.models.GaussianMixture`
        Desired state distribution.
    safety_factor : float
        Overbounding tuning factor for extra convergence term.
    y_ref : float
        Reference point to use on the sigmoid function, must be in the range
        (0, 1).

    Returns
    -------
    float
        cost.
    """
    all_goals = np.array([m.ravel().tolist() for m in goal_dist.means])
    all_states = np.array([m.ravel().tolist() for m in state_dist.means])
    target_center = np.mean(all_goals, axis=0).reshape((-1, 1))
    num_targets = all_goals.shape[0]
    num_objects = all_states.shape[0]
    state_dim = all_states.shape[1]

    # find radius of influence and shift
    diff = all_goals - target_center.T
    max_dist = np.sqrt(np.max(np.sum(diff * diff, axis=1)))
    radius_of_influence = safety_factor * max_dist
    shift = radius_of_influence + np.log(1 / y_ref - 1)

    # get actiavation term
    diff = all_states - target_center.T
    max_dist = np.sqrt(np.max(np.sum(diff * diff, axis=1)))
    activator = 1 / (1 + np.exp(-(max_dist - shift)))

    # get maximum variance
    max_var_obj = max(map(lambda x: float(np.max(np.diag(x))), state_dist.covariances))
    max_var_target = max(
        map(lambda x: float(np.max(np.diag(x))), goal_dist.covariances)
    )

    # Loop for all double summation terms
    sum_obj_obj = 0
    sum_obj_target = 0
    quad = 0
    for out_w, out_dist in state_dist:
        # create temporary gaussian object for calculations
        temp_gauss = smodels.Gaussian(mean=out_dist.mean)

        # object to object cost
        for in_w, in_dist in state_dist:
            temp_gauss.covariance = out_dist.covariance + in_dist.covariance
            sum_obj_obj += in_w * out_w * temp_gauss.pdf(in_dist.mean)

        # object to target and quadratic
        for tar_w, tar_dist in goal_dist:
            # object to target
            temp_gauss.covariance = out_dist.covariance + tar_dist.covariance
            sum_obj_target += tar_w * out_w * temp_gauss.pdf(tar_dist.mean)

            # quadratic
            diff = out_dist.mean - tar_dist.mean
            log_term = (
                np.log(
                    (2 * np.pi) ** (-0.5 * state_dim)
                    / np.sqrt(la.det(temp_gauss.covariance))
                )
                - 0.5 * diff.T @ la.inv(temp_gauss.covariance) @ diff
            )
            quad += out_w * tar_w * log_term.item()

    sum_target_target = 0
    for out_w, out_dist in goal_dist:
        temp_gauss = smodels.Gaussian(mean=out_dist.mean)
        for in_w, in_dist in goal_dist:
            temp_gauss.covariance = out_dist.covariance + in_dist.covariance
            sum_target_target += out_w * in_w * temp_gauss.pdf(in_dist.mean)

    return (
        10
        * num_objects
        * max_var_obj
        * (sum_obj_obj - 2 * max_var_target * num_targets * sum_obj_target)
        + sum_target_target
        + activator * quad
    )


class ELQR:
    """Implements the ELQR algorithm for swarms.

    Notes
    -----
    This follows :cite:`Thomas2021_RecedingHorizonExtendedLinearQuadraticRegulatorforRFSBasedSwarms`.

    Attributes
    ----------
    max_iters : int
        Maximum number of iterations to optimize.
    tol : float
        Relative tolerance for convergence.
    """

    def __init__(self, max_iters=1e3, tol=1e-4):
        """Initialize an object.

        Parameters
        ----------
        max_iters : int, optional
            Maximum number of iterations to optimize. The default is 1e3.
        tol : float, optional
            Relative tolerance for convergence. The default is 1e-4.
        """
        super().__init__()

        self.max_iters = int(max_iters)
        self.tol = tol
        self.end_dist = smodels.GaussianMixture()
        self.non_quad_weight = 1

        self._singleELQR = gcontrol.ELQR()
        self._elqr_lst = []
        self._start_covs = []
        self._start_weights = []
        self._time_vec = np.array([])
        self._cur_ind = None
        self._non_quad_fun = None

        self.__safety_factor = 1
        self.__y_ref = 0.5

    def get_state_dist(self, tt):
        """Calculates the current state distribution.

        Parameters
        ----------
        tt : float
            Current timestep.

        Returns
        -------
        :class:`serums.models.GaussianMixture`
            State distribution.
        """
        kk = int(np.argmin(np.abs(tt - self._time_vec)))
        means = [
            params["traj"][kk, :].copy().reshape((-1, 1)) for params in self._elqr_lst
        ]
        return smodels.GaussianMixture(
            means=means,
            covariances=[c.copy() for c in self._start_covs],
            weights=self._start_weights.copy(),
        )

    def non_quad_fun_factory(self):
        r"""Factory for creating the non-quadratic cost function.

        This should generate a function of the same form needed by the single
        agent controller, that is `time, state, control input, end state,
        is initial flag, is final flag, *args`. For this class, it implements
        a GM density based cost function plus an optional additional cost function
        of the same form. It returns a float for the non-quadratic part of
        the cost.

        Returns
        -------
        callable
            function to calculate cost.
        """

        def non_quadratic_fun(
            tt, state, ctrl_input, end_state, is_initial, is_final, *args
        ):
            state_dist = self.get_state_dist(tt)
            state_dist.remove_components(
                [self._cur_ind,]  # noqa
            )
            state_dist.add_components(
                state.reshape((-1, 1)),
                self._start_covs[self._cur_ind],
                self._start_weights[self._cur_ind],
            )
            cost = gaussian_density_cost(
                state_dist, self.end_dist, self.__safety_factor, self.__y_ref
            )
            if self._non_quad_fun is not None:
                cost += self.non_quad_weight * self._non_quad_fun(
                    tt, state, ctrl_input, end_state, is_initial, is_final, *args
                )
            return cost

        return non_quadratic_fun

    def set_cost_model(
        self,
        safety_factor=None,
        y_ref=None,
        non_quad_fun=None,
        quad_modifier=None,
        non_quad_weight=None,
    ):
        """Set the parameters for the cost non-quadratic function.

        This sets the parameters for the built-in non-quadratic cost function
        and allows specifying an additional non-quadratic function (see
        :meth:`gncpy.control.ELQR`) for the agents. Any additional non-quadratic
        terms must be set here instead of within the single control model directly
        due to the built-in multi-agent non-quadratic cost.

        Parameters
        ----------
        safety_factor : float, optional
            Additional scaling factor on the radius of influence. The default
            is None.
        y_ref : float, optional
            Reference point on the sigmoid function to use when determining
            th radius of influence. Should be between 0 and 1 (non-inclusive).
            The default is None.
        non_quad_fun : callable, optional
            Optional additional non-quadratic cost. Must have the same form as
            that of the single agent controller. See :meth:`gncpy.control.ELQR`.
            The default is None.
        quad_modifier : callable, optional
            Function to modify the quadratization process, see
            :meth:`gncpy.control.ELQR`. The default is None.
        non_quad_weight : float
            Additional scaling to be applied to the additional non_quadratic
            cost (the non built-in term). The default is None
        """
        if safety_factor is not None:
            self.__safety_factor = safety_factor

        if y_ref is not None:
            self.__y_ref = y_ref

        if non_quad_fun is not None:
            self._non_quad_fun = non_quad_fun

        if quad_modifier is not None:
            self._singleELQR.set_cost_model(
                quad_modifier=quad_modifier, skip_validity_check=True
            )

        if non_quad_weight is not None:
            self.non_quad_weight = non_quad_weight

    def set_control_model(self, singleELQR):
        """Sets the single agent control model used.

        Parameters
        ----------
        singleELQR : :class:`gncpy.control.ELQR`
            Single agent controller for generating trajectories.
        quad_modifier : callable, optional
            Modifing function for the quadratization. See
            :meth:`gncpy.control.ELQR.set_cost_model`. The default is None.
        """
        self._singleELQR = deepcopy(singleELQR)

    def find_end_state(self, tt, cur_ind, cur_state):
        """Finds the ending state for the given current state.

        Parameters
        ----------
        tt : float
            Current timestep
        cur_ind : int
            Current index into the state distribution
        cur_state : N x 1 numpy array
            Current state.

        Returns
        -------
        N x 1 numpy array
            Best ending state given the current state.
        """
        all_ends = np.vstack([m.ravel() for m in self.end_dist.means])
        diff = all_ends - cur_state.T
        min_ind = int(np.argmin(np.sum(diff * diff, axis=1)))

        return self.end_dist.means[min_ind]

    def init_elqr_lst(self, tt, start_dist):
        """Initialize the list of single agent ELQR controllers.

        Parameters
        ----------
        tt : float
            current time.
        start_dist : :class:`serums.models.GaussianMixture`
            Starting gaussian mixture.

        Returns
        -------
        num_timesteps : int
            total number of timesteps.
        """
        self._elqr_lst = []
        for ind, (w, dist) in enumerate(start_dist):
            p = {}
            p["elqr"] = deepcopy(self._singleELQR)
            end_state = np.zeros((dist.location.size, 1))
            p["old_cost"], num_timesteps, p["traj"], self._time_vec = p["elqr"].reset(
                tt, dist.location, end_state
            )
            self._elqr_lst.append(p)

        # reset with proper end state
        for ind, ((w, dist), p) in enumerate(zip(start_dist, self._elqr_lst)):
            end_state = self.find_end_state(tt, ind, dist.location)
            p["elqr"].reset(tt, dist.location, end_state)

        return num_timesteps

    def targets_to_wayareas(self, end_states):
        """Converts target locations to wayareas with automatic scaling.

        Performs a Principal Component Analysis (PCA) on the ending state
        locations to create a Gaussian Mixture.

        Parameters
        ----------
        end_states : Nt x N numpy array
            All possible ending states, one per row.

        Returns
        -------
        :class:`serums.models.GaussianMixture`
            Ending state distribution.
        """

        def find_principal_components(data):
            num_samps = data.shape[0]
            num_feats = data.shape[1]

            mean = np.sum(data, 0) / num_samps
            covars = np.zeros((num_feats, num_feats))
            for ii in range(0, num_feats):
                for jj in range(0, num_feats):
                    acc = 0
                    for samp in range(0, num_samps):
                        acc += (data[samp, ii] - mean[ii]) * (data[samp, jj] - mean[jj])
                    covars[ii, jj] = acc / num_samps
            (w, comps) = la.eig(covars)
            inds = np.argsort(w)[::-1]
            return comps[:, inds].T

        def find_largest_proj_dist(new_dirs, old_dirs):
            vals = np.zeros(new_dirs.shape[0])
            for ii in range(0, new_dirs.shape[1]):
                for jj in range(0, old_dirs.shape[1]):
                    proj = np.abs(new_dirs[:, [ii]].T @ old_dirs[:, [jj]])
                    if proj > vals[ii]:
                        vals[ii] = proj
            return vals

        thresh = 1e-2

        wayareas = smodels.GaussianMixture()
        all_ends = np.vstack([s.ravel() for s in end_states])
        aug_end_states = np.vstack((all_ends, np.mean(all_ends, axis=0)))

        directions = np.zeros(
            (aug_end_states.shape[1], aug_end_states.shape[0], aug_end_states.shape[0])
        )
        for ii, s_pt in enumerate(aug_end_states):
            for jj, e_pt in enumerate(aug_end_states):
                directions[:, ii, jj] = e_pt - s_pt

        weight = 1 / len(end_states)
        for wp_ind, center in enumerate(all_ends):
            sample_data = np.delete(aug_end_states, wp_ind, axis=0)
            sample_dirs = np.delete(directions[:, wp_ind, :].squeeze(), wp_ind, axis=1)
            comps = find_principal_components(sample_data)
            vals = find_largest_proj_dist(comps, sample_dirs)
            vals[vals <= thresh] = thresh

            cov = comps @ np.diag(vals) @ la.inv(comps)
            wayareas.add_components(center.reshape((-1, 1)), cov, weight)

        return wayareas

    def gen_final_traj(
        self,
        num_timesteps,
        start,
        elqr,
        state_args,
        ctrl_args,
        cost_args,
        inv_state_args,
        inv_ctrl_args,
    ):
        """Generates the final trajectory state and control trajectories.

        Parameters
        ----------
        num_timesteps : int
            total number of timesteps.
        start : N x 1 numpy array
            Initial state.
        elqr : :class:`gncpy.control.ELQR`
            Single agent controller for the given starting state.
        state_args : tuple
            Additional arguments for the state matrix.
        ctrl_args : tuple
            Additional arguments for the input matrix.
        cost_args : tuple
            Additional arguments for the cost function.
        inv_state_args : tuple
            Additional arguments for the inverse state transition matrix.
        inv_ctrl_args : tuple
            Additional arguments for the inverse input matrix.

        Returns
        -------
        state_traj : Nh+1 x N numpy array
            state trajectory.
        ctrl_signal : Nh x Nu numpy array
            control signal.
        cost : float
            cost of the trajectory.
        """
        ctrl_signal = np.nan * np.ones((num_timesteps, elqr.u_nom.size))
        state_traj = np.nan * np.ones((num_timesteps + 1, start.size))
        cost = 0
        state_traj[0, :] = start.flatten()
        for kk, tt in enumerate(self._time_vec[:-1]):
            ctrl_signal[kk, :] = (
                elqr.feedback_gain[kk] @ state_traj[kk, :].reshape((-1, 1))
                + elqr.feedthrough_gain[kk]
            ).ravel()
            if elqr.control_constraints is not None:
                ctrl_signal[kk, :] = elqr.control_constraints(
                    tt, ctrl_signal[kk, :].reshape((-1, 1))
                ).ravel()
            cost += elqr.cost_function(
                tt,
                state_traj[kk, :].reshape((-1, 1)),
                ctrl_signal[kk, :].reshape((-1, 1)),
                cost_args,
                is_initial=(kk == 0),
                is_final=False,
            )
            state_traj[kk + 1, :] = elqr.prop_state(
                tt,
                state_traj[kk, :].reshape((-1, 1)),
                ctrl_signal[kk, :].reshape((-1, 1)),
                state_args,
                ctrl_args,
                True,
                inv_state_args,
                inv_ctrl_args,
            ).ravel()

        cost += elqr.cost_function(
            self._time_vec[-1],
            state_traj[num_timesteps, :].reshape((-1, 1)),
            ctrl_signal[num_timesteps - 1, :].reshape((-1, 1)),
            cost_args,
            is_initial=False,
            is_final=True,
        )

        return state_traj, ctrl_signal, cost

    def draw_init_states(
        self, fig, states, plt_inds, marker, zorder, cmap=None, color=None
    ):
        kwargs = dict(marker=marker, zorder=zorder,)
        if color is not None:
            kwargs["color"] = color

        for c_ind, (w, dist) in enumerate(states):
            s = dist.location
            if cmap is not None:
                kwargs["color"] = cmap(c_ind)
            fig.axes[0].scatter(s[plt_inds[0], 0], s[plt_inds[1], 0], **kwargs)

    def save_animation(self, fig, fig_h, fig_w, frame_list):
        with io.BytesIO() as buff:
            fig.savefig(buff, format="raw")
            buff.seek(0)
            img = np.frombuffer(buff.getvalue(), dtype=np.uint8).reshape(
                (fig_h, fig_w, -1)
            )
        frame_list.append(Image.fromarray(img))

    def init_plot(
        self,
        show_animation,
        save_animation,
        cmap,
        start_dist,
        fig,
        plt_opts,
        ttl,
        plt_inds,
    ):
        frame_list = []
        fig_h = None
        fig_w = None
        if show_animation:
            if cmap is None:
                cmap = gplot.get_cmap(len(start_dist))

            if fig is None:
                fig = plt.figure()
                fig.add_subplot(1, 1, 1)
                fig.axes[0].set_aspect("equal", adjustable="box")

                if plt_opts is None:
                    plt_opts = gplot.init_plotting_opts(f_hndl=fig)

                if ttl is None:
                    ttl = "Multi-Agent ELQR"

                gplot.set_title_label(fig, 0, plt_opts, ttl=ttl)

                # draw start
                self.draw_init_states(fig, start_dist, plt_inds, "o", 1000, cmap=cmap)

            self.draw_init_states(fig, self.end_dist, plt_inds, "x", 1000, color="r")

            fig.tight_layout()
            plt.pause(0.1)

            # for stopping simulation with the esc key.
            fig.canvas.mpl_connect(
                "key_release_event",
                lambda event: [exit(0) if event.key == "escape" else None],
            )
            fig_w, fig_h = fig.canvas.get_width_height()

            # save first frame of animation
            if save_animation:
                self.save_animation(fig, fig_h, fig_w, frame_list)

        return fig, fig_h, fig_w, frame_list, cmap

    def reset(self, start_dist):
        self._start_covs = [c.copy() for c in start_dist.covariances]
        self._start_weights = [w for w in start_dist.weights]

    def output_helper(
        self,
        c_ind,
        num_timesteps,
        start,
        params,
        state_args,
        ctrl_args,
        cost_args,
        inv_state_args,
        inv_ctrl_args,
        costs,
        state_trajs,
        ctrl_signals,
        show_animation,
        fig,
        plt_inds,
        cmap,
    ):
        self._cur_ind = c_ind
        params["elqr"].set_cost_model(
            non_quadratic_fun=self.non_quad_fun_factory(), skip_validity_check=True
        )
        st, cs, c = self.gen_final_traj(
            num_timesteps,
            start,
            params["elqr"],
            state_args,
            ctrl_args,
            cost_args,
            inv_state_args,
            inv_ctrl_args,
        )
        costs.append(c)
        state_trajs.append(st)
        ctrl_signals.append(cs)

        if show_animation:
            fig.axes[0].plot(
                st[:, plt_inds[0]],
                st[:, plt_inds[1]],
                linestyle="-",
                color=cmap(c_ind),
            )
            plt.pause(0.01)

    def create_outputs(
        self,
        start_dist,
        num_timesteps,
        state_args,
        ctrl_args,
        cost_args,
        inv_state_args,
        inv_ctrl_args,
        show_animation,
        fig,
        plt_inds,
        cmap,
    ):
        costs = []
        state_trajs = []
        ctrl_signals = []
        for c_ind, ((w, dist), params) in enumerate(zip(start_dist, self._elqr_lst)):
            self.output_helper(
                c_ind,
                num_timesteps,
                dist.location,
                params,
                state_args,
                ctrl_args,
                cost_args,
                inv_state_args,
                inv_ctrl_args,
                costs,
                state_trajs,
                ctrl_signals,
                show_animation,
                fig,
                plt_inds,
                cmap,
            )

        return state_trajs, costs, ctrl_signals

    def plan(
        self,
        tt,
        start_dist,
        end_dist,
        state_args=None,
        ctrl_args=None,
        cost_args=None,
        inv_state_args=None,
        inv_ctrl_args=None,
        provide_details=False,
        disp=True,
        show_animation=False,
        save_animation=False,
        plt_opts=None,
        ttl=None,
        fig=None,
        cmap=None,
        plt_inds=None,
        update_end=True,
        update_end_iters=1,
    ):
        """Main planning function.

        Parameters
        ----------
        tt : float
            Starting timestep for the plan.
        start_dist : :class:`serums.models.GaussianMixture`
            Starting state distribution.
        end_dist : :class:`serums.models.GaussianMixture`
            Ending state distribution.
        state_args : tuple, optional
            Additional arguments for getting the state transition matrix. The
            default is None.
        ctrl_args : tuple, optional
            Additional arguements for getting the input matrix. The default is
            None.
        cost_args : tuple, optional
            Additional arguments for the cost function. The default is None.
        inv_state_args : tuple, optional
            Additional arguments to get the inverse state matrix. The default
            is None.
        inv_ctrl_args : tuple, optional
            Additional arguments to get the inverse input matrix. The default
            is None.
        provide_details : bool, optional
            Falg for if optional outputs should be output. The default is False.
        disp : bool, optional
            Falg for if additional text should be printed. The default is True.
        show_animation : bool, optional
            Flag for if an animation is generated. The default is False.
        save_animation : bool, optional
            Flag for saving the animation. Only applies if the animation is
            shown. The default is False.
        plt_opts : dict, optional
            Additional plotting options. See
            :func:`gncpy.plotting.init_plotting_opts`. The default is None.
        ttl : string, optional
            Title of the generated plot. The default is None.
        fig : matplotlib figure, optional
            Handle to the figure. If supplied only the end states are added.
            The default is None.
        cmap : matplotlib colormap, optional
            Color map for the different agents. See :func:`gncpy.plotting.get_cmap`.
            The default is None.
        plt_inds : list, optional
            Indices in the state vector to plot. The default is None.

        Returns
        -------
        state_trajs : list
            Each element is an Nh+1xN numpy array.
        costs : list, optional
            Each element is a float for the cost of that trajectory
        ctrl_signals : list, optional
            Each element is an NhxNu numpy array
        fig : matplotlib figure, optional
            Handle to the generated figure
        frame_list : list, optional
            Each element is a PIL image if the animation is being saved.
        """
        if state_args is None:
            state_args = ()
        if ctrl_args is None:
            ctrl_args = ()
        if cost_args is None:
            cost_args = ()
        if inv_state_args is None:
            inv_state_args = ()
        if inv_ctrl_args is None:
            inv_ctrl_args = ()
        if plt_inds is None:
            plt_inds = [0, 1]

        self.end_dist = end_dist

        num_timesteps = self.init_elqr_lst(tt, start_dist)
        old_cost = float("inf")
        self.reset(start_dist)

        fig, fig_h, fig_w, frame_list, cmap = self.init_plot(
            show_animation,
            save_animation,
            cmap,
            start_dist,
            fig,
            plt_opts,
            ttl,
            plt_inds,
        )

        if disp:
            print("Starting ELQR optimization loop...")

        for itr in range(self.max_iters):
            # forward pass for each gaussian, step by step
            for kk in range(num_timesteps):
                for ind, params in enumerate(self._elqr_lst):
                    self._cur_ind = ind
                    params["elqr"].set_cost_model(
                        non_quadratic_fun=self.non_quad_fun_factory(),
                        skip_validity_check=True,
                    )
                    params["traj"][kk + 1, :] = params["elqr"].forward_pass_step(
                        itr,
                        kk,
                        self._time_vec,
                        params["traj"],
                        state_args,
                        ctrl_args,
                        cost_args,
                        inv_state_args,
                        inv_ctrl_args,
                    )

            # quadratize final cost for each gaussian
            for c_ind, params in enumerate(self._elqr_lst):
                self._cur_ind = c_ind
                params["elqr"].set_cost_model(
                    non_quadratic_fun=self.non_quad_fun_factory(),
                    skip_validity_check=True,
                )
                # if update_end and itr != 0 and itr % update_end_iters == 0:
                #     params["elqr"].end_state = self.find_end_state(
                #         self._time_vec[-1], c_ind, params["traj"][-1, :].reshape((-1, 1)),
                #     )
                params["traj"] = params["elqr"].quadratize_final_cost(
                    itr, num_timesteps, params["traj"], self._time_vec, cost_args
                )

            # backward pass for each gaussian
            for kk in range(num_timesteps - 1, -1, -1):
                for c_ind, params in enumerate(self._elqr_lst):
                    self._cur_ind = c_ind
                    params["elqr"].set_cost_model(
                        non_quadratic_fun=self.non_quad_fun_factory(),
                        skip_validity_check=True,
                    )
                    params["traj"][kk, :] = params["elqr"].backward_pass_step(
                        itr,
                        kk,
                        self._time_vec,
                        params["traj"],
                        state_args,
                        ctrl_args,
                        cost_args,
                        inv_state_args,
                        inv_ctrl_args,
                    )

            # get true cost
            cost = 0
            x = np.nan * np.ones(
                (len(self._elqr_lst), self._elqr_lst[0]["traj"].shape[1])
            )
            for kk, tt in enumerate(self._time_vec[:-1]):
                for ind, params in enumerate(self._elqr_lst):
                    x[ind] = params["traj"][kk].copy()
                    u = (
                        params["elqr"].feedback_gain[kk] @ x[ind].reshape((-1, 1))
                        + params["elqr"].feedthrough_gain[kk]
                    )
                    if params["elqr"].control_constraints is not None:
                        u = params["elqr"].control_constraints(tt, u)
                    self._cur_ind = ind
                    params["elqr"].set_cost_model(
                        non_quadratic_fun=self.non_quad_fun_factory(),
                        skip_validity_check=True,
                    )
                    cost += params["elqr"].cost_function(
                        tt,
                        x[ind].reshape((-1, 1)),
                        u,
                        cost_args,
                        is_initial=(kk == 0),
                        is_final=False,
                    )
                    x[ind] = (
                        params["elqr"]
                        .prop_state(
                            tt,
                            x[ind].reshape((-1, 1)),
                            u,
                            state_args,
                            ctrl_args,
                            True,
                            inv_state_args,
                            inv_ctrl_args,
                        )
                        .ravel()
                    )

            for ind, params in enumerate(self._elqr_lst):
                if update_end and itr != 0 and itr % update_end_iters == 0:
                    params["elqr"].end_state = self.find_end_state(
                        self._time_vec[-1], ind, x[ind].reshape((-1, 1))
                    )
                u = (
                    params["elqr"].feedback_gain[-1] @ x[ind].reshape((-1, 1))
                    + params["elqr"].feedthrough_gain[-1]
                )
                if params["elqr"].control_constraints is not None:
                    u = params["elqr"].control_constraints(tt, u)

                self._cur_ind = ind
                params["elqr"].set_cost_model(
                    non_quadratic_fun=self.non_quad_fun_factory(),
                    skip_validity_check=True,
                )
                cost += params["elqr"].cost_function(
                    self._time_vec[-1],
                    x[ind].reshape((-1, 1)),
                    u,
                    cost_args,
                    is_initial=False,
                    is_final=True,
                )

            # for ind, params in enumerate(self._elqr_lst):
            #     self._cur_ind = ind
            #     params["elqr"].set_cost_model(
            #         non_quadratic_fun=self.non_quad_fun_factory(),
            #         skip_validity_check=True,
            #     )
            #     x = params["traj"][0, :].copy().reshape((-1, 1))
            #     for kk, tt in enumerate(self._time_vec[:-1]):
            #         u = (
            #             params["elqr"].feedback_gain[kk] @ x
            #             + params["elqr"].feedthrough_gain[kk]
            #         )
            #         if params["elqr"].control_constraints is not None:
            #             u = params["elqr"].control_constraints(tt, u)
            #         cost += params["elqr"].cost_function(
            #             tt, x, u, cost_args, is_initial=(kk == 0), is_final=False,
            #         )
            #         x = params["elqr"].prop_state(
            #             tt,
            #             x,
            #             u,
            #             state_args,
            #             ctrl_args,
            #             True,
            #             inv_state_args,
            #             inv_ctrl_args,
            #         )
            #     params["elqr"].end_state = self.find_end_state(
            #         self._time_vec[-1], ind, params["traj"][-1, :].reshape((-1, 1))
            #     )
            #     cost += params["elqr"].cost_function(
            #         self._time_vec[-1],
            #         x,
            #         u,
            #         cost_args,
            #         is_initial=False,
            #         is_final=True,
            #     )

            if disp:
                print("\tIteration: {:3d} Cost: {:10.4f}".format(itr, cost))

            if show_animation:
                for c_ind, params in enumerate(self._elqr_lst):
                    img = params["elqr"].draw_traj(
                        fig,
                        plt_inds,
                        fig_h,
                        fig_w,
                        c_ind == (len(self._elqr_lst) - 1),
                        num_timesteps,
                        self._time_vec,
                        state_args,
                        ctrl_args,
                        inv_state_args,
                        inv_ctrl_args,
                        color=cmap(c_ind),
                        alpha=0.2,
                        zorder=-10,
                    )

                if save_animation:
                    frame_list.append(img)

            # check for convergence
            if np.abs((old_cost - cost) / cost) < self.tol:
                break
            old_cost = cost

        # generate control and state trajectories for all agents
        state_trajs, costs, ctrl_signals = self.create_outputs(
            start_dist,
            num_timesteps,
            state_args,
            ctrl_args,
            cost_args,
            inv_state_args,
            inv_ctrl_args,
            show_animation,
            fig,
            plt_inds,
            cmap,
        )

        if show_animation and save_animation:
            plt.pause(0.01)
            self.save_animation(fig, fig_h, fig_w, frame_list)

        details = (costs, ctrl_signals, fig, frame_list)
        return (state_trajs, *details) if provide_details else state_trajs


class ELQROSPA(ELQR):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.end_dist = np.array([])

        self.loiter_dist = 1

        self.__ospa_inds = []
        self.__ospa_cutoff = 1e3
        self.__ospa_core = SingleObjectDistance.EUCLIDEAN

    def get_state_dist(self, tt):
        """Calculates the current state distribution.

        Parameters
        ----------
        tt : float
            Current timestep.

        Returns
        -------
        Na x N numpy array
            State distribution, one row per agent
        """
        kk = int(np.argmin(np.abs(tt - self._time_vec)))
        return np.vstack([params["traj"][kk, :].flatten() for params in self._elqr_lst])

    def non_quad_fun_factory(self):
        """Factory for creating the non-quadratic cost function.

        This should generate a function of the same form needed by the single
        agent controller, that is `time, state, control input, end state,
        is initial flag, is final flag, *args`. For this class, it implements
        an OSPA based cost function plus an optional additional cost function
        of the same form. It returns a float for the non-quadratic part of
        the cost.

        Returns
        -------
        callable
            function to calculate cost.
        """

        def non_quadratic_fun(
            tt, state, ctrl_input, end_state, is_initial, is_final, *args
        ):
            if self.__ospa_inds is None:
                inds = [i for i in range(state.size)]
            else:
                inds = self.__ospa_inds
            start_dist = self.get_state_dist(tt)
            start_dist[self._cur_ind, :] = state.flatten()
            start_dist = start_dist[:, inds].T.reshape(
                (len(inds), 1, start_dist.shape[0])
            )

            end_dist = self.end_dist[:, inds].T.reshape(
                (len(inds), 1, self.end_dist.shape[0])
            )

            cost = calculate_ospa(
                start_dist,
                end_dist,
                self.__ospa_cutoff,
                1,
                core_method=self.__ospa_core,
            )[0].item()
            if self._non_quad_fun is not None:
                cost += self.non_quad_weight * self._non_quad_fun(
                    tt, state, ctrl_input, end_state, is_initial, is_final, *args
                )
            return cost

        return non_quadratic_fun

    def set_cost_model(
        self,
        ospa_inds=None,
        ospa_cutoff=None,
        ospa_method=None,
        non_quad_fun=None,
        quad_modifier=None,
        non_quad_weight=None,
    ):
        """Set the parameters for the cost non-quadratic function.

        This sets the parameters for the built-in non-quadratic cost function
        and allows specifying an additional non-quadratic function (see
        :meth:`gncpy.control.ELQR`) for the agents. Any additional non-quadratic
        terms must be set here instead of within the single control model directly
        due to the built-in multi-agent non-quadratic cost.

        Parameters
        ----------
        non_quad_fun : callable, optional
            Optional additional non-quadratic cost. Must have the same form as
            that of the single agent controller. See :meth:`gncpy.control.ELQR`.
            The default is None.
        quad_modifier : callable, optional
            Function to modify the quadratization process, see
            :meth:`gncpy.control.ELQR`. The default is None.
        non_quad_weight : float
            Additional scaling to be applied to the additional non_quadratic
            cost (the non built-in term). The default is None
        """
        if ospa_inds is not None:
            self.__ospa_inds = ospa_inds

        if ospa_cutoff is not None:
            self.__ospa_cutoff = ospa_cutoff

        if ospa_method is not None:
            self.__ospa_core = ospa_method

        if non_quad_fun is not None:
            self._non_quad_fun = non_quad_fun

        if quad_modifier is not None:
            self._singleELQR.set_cost_model(
                quad_modifier=quad_modifier, skip_validity_check=True
            )

        if non_quad_weight is not None:
            self.non_quad_weight = non_quad_weight

    def find_end_state(self, tt, cur_ind, cur_state):
        """Finds the ending state for the given current state.

        Parameters
        ----------
        tt : float
            Current timestep
        cur_ind : int
            Current index into the state distribution
        cur_state : N x 1 numpy array
            Current state.

        Returns
        -------
        N x 1 numpy array
            Best ending state given the current state.
        """
        if self.__ospa_inds is None:
            inds = [i for i in range(cur_state.size)]
        else:
            inds = self.__ospa_inds
        dist = self.get_state_dist(tt)
        dist[cur_ind] = cur_state.ravel()
        dist = dist[:, inds]

        end_dist = self.end_dist[:, inds]
        if end_dist.shape[0] < dist.shape[0]:
            if end_dist.shape[0] > 1:
                center = np.mean(end_dist, axis=0).reshape((1, -1))
            else:
                # direction from end to current state
                direc = cur_state.ravel()[inds] - end_dist[0]
                mag = np.linalg.norm(direc)
                if mag > 1e-16:
                    direc /= mag
                else:
                    # current agent is at the only target, use all agents center instead
                    a_cent = np.mean(dist, axis=0)
                    direc = a_cent - end_dist[0]
                    mag = np.linalg.norm(direc)
                    if mag > 1e-16:
                        direc /= mag
                    else:
                        # all agents center is at the only target...just pick x direction
                        direc = np.zeros(len(inds))
                        direc[0] = 1
                direc *= self.loiter_dist
                center = (end_dist[0] + direc).reshape((1, -1))

            # add enough "fake" targets so there is a 1-to-1 matching
            n_miss = dist.shape[0] - end_dist.shape[0]
            end_dist = np.vstack((end_dist, center * np.ones((n_miss, len(inds)))))

        distances, a_exists, t_exists = calculate_ospa(
            dist.T.reshape((len(inds), 1, -1)),
            end_dist.T.reshape((len(inds), 1, -1)),
            self.__ospa_cutoff,
            1,
            core_method=self.__ospa_core,
            use_empty=False,
        )[6:]
        cont_sub = distances[
            0 : np.sum(a_exists).astype(int), 0 : np.sum(t_exists).astype(int), 0
        ]
        tar_inds = linear_sum_assignment(cont_sub)[1]

        min_ind = tar_inds[cur_ind]
        if min_ind >= self.end_dist.shape[0]:
            out = np.zeros((self.end_dist.shape[1], 1))
            out[inds] = end_dist[min_ind].reshape(out[inds].shape)
        else:
            out = self.end_dist[min_ind, :].reshape((-1, 1))
        return out

    def init_elqr_lst(self, tt, start_dist):
        """Initialize the list of single agent ELQR controllers.

        Parameters
        ----------
        tt : float
            current time.
        start_dist : Na x N numpy array
            Starting states, one per row.

        Returns
        -------
        num_timesteps : int
            total number of timesteps.
        """
        self._elqr_lst = []
        # initialize with fake end state
        for ind, s in enumerate(start_dist):
            p = {}
            p["elqr"] = deepcopy(self._singleELQR)
            end_state = np.zeros((s.size, 1))
            p["old_cost"], num_timesteps, p["traj"], self._time_vec = p["elqr"].reset(
                tt, s.reshape((-1, 1)), end_state
            )
            self._elqr_lst.append(p)

        # reset with proper end state
        for ind, (s, p) in enumerate(zip(start_dist, self._elqr_lst)):
            end_state = self.find_end_state(tt, ind, s)
            p["elqr"].reset(tt, s.reshape((-1, 1)), end_state)

        return num_timesteps

    def targets_to_wayareas(self, end_states):
        warn("targets_to_wayareas not used by ELQROSPA")
        return None

    def draw_init_states(
        self, fig, states, plt_inds, marker, zorder, color=None, cmap=None
    ):
        kwargs = dict(marker=marker, zorder=zorder,)
        if color is not None:
            kwargs["color"] = color

        for c_ind, s in enumerate(states):
            if cmap is not None:
                kwargs["color"] = cmap(c_ind)
            fig.axes[0].scatter(s[plt_inds[0]], s[plt_inds[1]], **kwargs)

    def reset(self, start_dist):
        pass

    def create_outputs(
        self,
        start_dist,
        num_timesteps,
        state_args,
        ctrl_args,
        cost_args,
        inv_state_args,
        inv_ctrl_args,
        show_animation,
        fig,
        plt_inds,
        cmap,
    ):
        costs = []
        state_trajs = []
        ctrl_signals = []
        for c_ind, (s, params) in enumerate(zip(start_dist, self._elqr_lst)):
            self.output_helper(
                c_ind,
                num_timesteps,
                s.reshape((-1, 1)),
                params,
                state_args,
                ctrl_args,
                cost_args,
                inv_state_args,
                inv_ctrl_args,
                costs,
                state_trajs,
                ctrl_signals,
                show_animation,
                fig,
                plt_inds,
                cmap,
            )

        return state_trajs, costs, ctrl_signals

    def plan(self, tt, start_dist, end_dist, **kwargs):
        return super().plan(tt, start_dist, end_dist, **kwargs)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from importlib.metadata import version as get_version

sys.path.append(os.path.abspath("../../carbs"))

sys.path.append(os.path.abspath("."))
from example_runner import run_examples

run_examples()

# -- Project information -----------------------------------------------------

project = "CARBS"
copyright = "2019, Jordan D Larson"
author = "LAGER"
version = get_version("carbs")
release = get_version("carbs")


# -- General configuration ---------------------------------------------------
# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinxcontrib.bibtex",
    "sphinx_sitemap",
    "sphinx_copybutton",
    "sphinx_rtd_theme",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
# exclude_patterns = ['_templates']

# Configuration for autodoc/summary
# note see https://stackoverflow.com/questions/2701998/sphinx-autodoc-is-not-automatic-enough
# for help/details
autosummary_generate = True
autodoc_member_order = "groupwise"
add_module_names = False
modindex_common_prefix = [
    "carbs.",
]

# configure copy button for code snippets
copybutton_only_copy_prompt_lines = False

# Todo configuration
todo_include_todos = True
todo_link_only = True

# bibtex config
bibtex_bibfiles = ["refs.bib"]


# -- Options for HTML output -------------------------------------------------
html_theme = "sphinx_rtd_theme"
html_theme_options = {
    "display_version": True,
    "style_nav_header_background": "#9E1B32",
    # Toc options
    "collapse_navigation": False,
    "sticky_navigation": True,
    "navigation_depth": 4,
}
html_show_sourcelink = False
html_baseurl = "https://drjdlarson.github.io/carbs/"
html_extra_path = [
    "robots.txt",
]  # robots.txt is for search engine stuff
html_logo = "logo.svg"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']
